@@ -669,40 +669,85 @@ def _fix_encode_to_binary(tokens, i):  # type: (List[Token], int) -> None
     del tokens[victims]
 
 
-FUTURES = {}  # type: Dict[Tuple[int, ...], Tuple[str, ...]]
-FUTURES[(2, 7)] = ('nested_scopes', 'generators', 'with_statement')
-FUTURES[(3,)] = FUTURES[(3, 6)] = FUTURES[(2, 7)] + (
-    'absolute_import', 'division', 'print_function', 'unicode_literals',
-)
-FUTURES[(3, 7)] = FUTURES[(3, 6)] + ('generator_stop',)
+def _build_import_removals():
+    # type: () -> Dict[Tuple[int, ...], Dict[str, Tuple[str, ...]]]
+    ret = {}
+    future = (
+        ((2, 7), ('nested_scopes', 'generators', 'with_statement')),
+        (
+            (3,), (
+                'absolute_import', 'division', 'print_function',
+                'unicode_literals',
+            ),
+        ),
+        ((3, 6), ()),
+        ((3, 7), ('generator_stop',)),
+    )  # type: Tuple[Tuple[Tuple[int, ...], Tuple[str, ...]], ...]
+
+    prev = ()  # type: Tuple[str, ...]
+    for min_version, names in future:
+        prev += names
+        ret[min_version] = {'__future__': names}
+    # see reorder_python_imports
+    for k, v in ret.items():
+        if k >= (3,):
+            v.update({
+                'builtins': (
+                    'ascii', 'bytes', 'chr', 'dict', 'filter', 'hex', 'input',
+                    'int', 'list', 'map', 'max', 'min', 'next', 'object',
+                    'oct', 'open', 'pow', 'range', 'round', 'str', 'super',
+                    'zip', '*',
+                ),
+                'io': ('open',),
+                'six': ('callable', 'next'),
+                'six.moves': ('filter', 'input', 'map', 'range', 'zip'),
+            })
+    return ret
 
 
-def _fix_future_imports(tokens, start, min_version):
+IMPORT_REMOVALS = _build_import_removals()
+
+
+def _fix_import_removals(tokens, start, min_version):
     # type: (List[Token], int, Tuple[int, ...]) -> None
     i = start + 1
-    while tokens[i].name != 'NAME':
+    name_parts = []
+    while tokens[i].src != 'import':
+        if tokens[i].name in {'NAME', 'OP'}:
+            name_parts.append(tokens[i].src)
         i += 1
-    if tokens[i].src != '__future__':
+
+    modname = ''.join(name_parts)
+    if modname not in IMPORT_REMOVALS[min_version]:
         return
 
-    found_names = []  # type: List[int]
+    found = []  # type: List[Optional[int]]
     i += 1
     while tokens[i].name not in {'NEWLINE', 'ENDMARKER'}:
-        if tokens[i].name == 'NAME' and tokens[i].src != 'import':
-            found_names.append(i)
+        if tokens[i].name == 'NAME' or tokens[i].src == '*':
+            # don't touch aliases
+            if (
+                    found and found[-1] is not None and
+                    tokens[found[-1]].src == 'as'
+            ):
+                found[-2:] = [None]
+            else:
+                found.append(i)
         i += 1
     # depending on the version of python, some will not emit NEWLINE('') at the
     # end of a file which does not end with a newline (for example 2.7.6)
     if tokens[i].name == 'ENDMARKER':  # pragma: no cover
         i -= 1
 
-    remove_futures = FUTURES[min_version]
-    to_remove = [x for x in found_names if tokens[x].src in remove_futures]
-    if len(to_remove) == len(found_names):
+    remove_names = IMPORT_REMOVALS[min_version][modname]
+    to_remove = [
+        x for x in found if x is not None and tokens[x].src in remove_names
+    ]
+    if len(to_remove) == len(found):
         del tokens[start:i + 1]
     else:
         for idx in reversed(to_remove):
-            if found_names[0] == idx:  # look forward until next name and del
+            if found[0] == idx:  # look forward until next name and del
                 j = idx + 1
                 while tokens[j].name != 'NAME':
                     j += 1
@@ -747,7 +792,7 @@ def _fix_tokens(contents_text, min_version):
             if tokens[i].name == 'NL':  # pragma: no branch (old PY2)
                 del tokens[i]
         elif token.src == 'from' and token.utf8_byte_offset == 0:
-            _fix_future_imports(tokens, i, min_version)
+            _fix_import_removals(tokens, i, min_version)
     return tokens_to_src(tokens).lstrip()
 
 
@@ -1170,8 +1215,9 @@ class FindPy3Plus(ast.NodeVisitor):
         self.encode_calls = {}  # type: Dict[Offset, ast.Call]
 
         self._version_info_imported = False
-        self.if_py2_blocks = set()  # type: Set[Offset]
         self.if_py3_blocks = set()  # type: Set[Offset]
+        self.if_py2_blocks_else = set()  # type: Set[Offset]
+        self.if_py3_blocks_else = set()  # type: Set[Offset]
 
         self.native_literals = set()  # type: Set[Offset]
 
@@ -1383,7 +1429,7 @@ class FindPy3Plus(ast.NodeVisitor):
             # _is_six() enforces this
             assert isinstance(arg, (ast.Name, ast.Attribute))
             self.six_type_ctx[_ast_to_offset(node.args[1])] = arg
-        elif self._is_six(node.func, ('b',)):
+        elif self._is_six(node.func, ('b', 'ensure_binary')):
             self.six_b.add(_ast_to_offset(node))
         elif self._is_six(node.func, SIX_CALLS) and not _starargs(node):
             self.six_calls[_ast_to_offset(node)] = node
@@ -1420,8 +1466,10 @@ class FindPy3Plus(ast.NodeVisitor):
         ):
             self.super_calls[_ast_to_offset(node)] = node
         elif (
-                isinstance(node.func, ast.Name) and
-                node.func.id == 'str' and
+                (
+                    self._is_six(node.func, ('ensure_str', 'ensure_text')) or
+                    isinstance(node.func, ast.Name) and node.func.id == 'str'
+                ) and
                 not node.keywords and
                 not _starargs(node) and
                 (
@@ -1476,46 +1524,49 @@ class FindPy3Plus(ast.NodeVisitor):
         return elts[0].n == 3 and all(n.n == 0 for n in elts[1:])
 
     def visit_If(self, node):  # type: (ast.If) -> None
-        if node.orelse and not isinstance(node.orelse[0], ast.If):
-            if (
-                    # if six.PY2:
-                    self._is_six(node.test, ('PY2',)) or
-                    # if not six.PY3:
-                    (
-                        isinstance(node.test, ast.UnaryOp) and
-                        isinstance(node.test.op, ast.Not) and
-                        self._is_six(node.test.operand, ('PY3',))
-                    ) or
-                    # sys.version_info == 2 or < (3,)
-                    (
-                        isinstance(node.test, ast.Compare) and
-                        self._is_version_info(node.test.left) and
-                        len(node.test.ops) == 1 and (
-                            self._eq(node.test, 2) or
-                            self._compare_to_3(node.test, ast.Lt)
-                        )
+        if (
+                # if six.PY2:
+                self._is_six(node.test, ('PY2',)) or
+                # if not six.PY3:
+                (
+                    isinstance(node.test, ast.UnaryOp) and
+                    isinstance(node.test.op, ast.Not) and
+                    self._is_six(node.test.operand, ('PY3',))
+                ) or
+                # sys.version_info == 2 or < (3,)
+                (
+                    isinstance(node.test, ast.Compare) and
+                    self._is_version_info(node.test.left) and
+                    len(node.test.ops) == 1 and (
+                        self._eq(node.test, 2) or
+                        self._compare_to_3(node.test, ast.Lt)
                     )
-            ):
-                self.if_py2_blocks.add(_ast_to_offset(node))
-            elif (
-                    # if six.PY3:
-                    self._is_six(node.test, 'PY3') or
-                    # if not six.PY2:
-                    (
-                        isinstance(node.test, ast.UnaryOp) and
-                        isinstance(node.test.op, ast.Not) and
-                        self._is_six(node.test.operand, ('PY2',))
-                    ) or
-                    # sys.version_info == 3 or >= (3,) or > (3,)
-                    (
-                        isinstance(node.test, ast.Compare) and
-                        self._is_version_info(node.test.left) and
-                        len(node.test.ops) == 1 and (
-                            self._eq(node.test, 3) or
-                            self._compare_to_3(node.test, (ast.Gt, ast.GtE))
-                        )
+                )
+        ):
+            if node.orelse and not isinstance(node.orelse[0], ast.If):
+                self.if_py2_blocks_else.add(_ast_to_offset(node))
+        elif (
+                # if six.PY3:
+                self._is_six(node.test, 'PY3') or
+                # if not six.PY2:
+                (
+                    isinstance(node.test, ast.UnaryOp) and
+                    isinstance(node.test.op, ast.Not) and
+                    self._is_six(node.test.operand, ('PY2',))
+                ) or
+                # sys.version_info == 3 or >= (3,) or > (3,)
+                (
+                    isinstance(node.test, ast.Compare) and
+                    self._is_version_info(node.test.left) and
+                    len(node.test.ops) == 1 and (
+                        self._eq(node.test, 3) or
+                        self._compare_to_3(node.test, (ast.Gt, ast.GtE))
                     )
-            ):
+                )
+        ):
+            if node.orelse and not isinstance(node.orelse[0], ast.If):
+                self.if_py3_blocks_else.add(_ast_to_offset(node))
+            elif not node.orelse:
                 self.if_py3_blocks.add(_ast_to_offset(node))
         self.generic_visit(node)
 
@@ -1565,7 +1616,9 @@ def _find_block_start(tokens, i):  # type: (List[Token], int) -> int
 
 
 class Block(
-        collections.namedtuple('Block', ('start', 'block', 'end', 'line')),
+        collections.namedtuple(
+            'Block', ('start', 'colon', 'block', 'end', 'line'),
+        ),
 ):
     __slots__ = ()
 
@@ -1602,6 +1655,10 @@ class Block(
             ):
                 tokens[i] = tokens[i]._replace(src=tokens[i].src[diff:])
 
+    def replace_condition(self, tokens, new):
+        # type: (List[Token], List[Token]) -> None
+        tokens[self.start:self.colon] = new
+
     def _trim_end(self, tokens):  # type: (List[Token]) -> Block
         """the tokenizer reports the end of the block at the beginning of
         the next block
@@ -1645,10 +1702,11 @@ class Block(
             while level:
                 level += {'INDENT': 1, 'DEDENT': -1}.get(tokens[j].name, 0)
                 j += 1
+            ret = cls(start, colon, block, j, line=False)
             if trim_end:
-                return cls(start, block, j, line=False)._trim_end(tokens)
+                return ret._trim_end(tokens)
             else:
-                return cls(start, block, j, line=False)
+                return ret
         else:  # single line block
             block = j
             # search forward until the NEWLINE token
@@ -1657,7 +1715,7 @@ class Block(
             # we also want to include the newline in the block
             if tokens[j].name == 'NEWLINE':  # pragma: no branch (PY2 only)
                 j += 1
-            return cls(start, block, j, line=True)
+            return cls(start, colon, block, j, line=True)
 
 
 def _find_if_else_block(tokens, i):
@@ -1670,6 +1728,13 @@ def _find_if_else_block(tokens, i):
     return if_block, else_block
 
 
+def _find_elif(tokens, i):
+    # type: (List[Token], int) -> int
+    while tokens[i].src != 'elif':  # pragma: no cover (only for <3.8.1)
+        i -= 1
+    return i
+
+
 def _remove_decorator(tokens, i):  # type: (List[Token], int) -> None
     while tokens[i - 1].src != '@':
         i -= 1
@@ -1805,8 +1870,9 @@ def _fix_py3_plus(contents_text):  # type: (str) -> str
     if not any((
             visitor.bases_to_remove,
             visitor.encode_calls,
-            visitor.if_py2_blocks,
+            visitor.if_py2_blocks_else,
             visitor.if_py3_blocks,
+            visitor.if_py3_blocks_else,
             visitor.native_literals,
             visitor.io_open_calls,
             visitor.os_error_alias_calls,
@@ -1852,21 +1918,34 @@ def _fix_py3_plus(contents_text):  # type: (str) -> str
             continue
         elif token.offset in visitor.bases_to_remove:
             _remove_base_class(tokens, i)
-        elif token.offset in visitor.if_py2_blocks:
-            if tokens[i].src != 'if':
-                continue
-            if_block, else_block = _find_if_else_block(tokens, i)
-
-            else_block.dedent(tokens)
-            del tokens[if_block.start:else_block.block]
         elif token.offset in visitor.if_py3_blocks:
-            if tokens[i].src != 'if':
-                continue
-            if_block, else_block = _find_if_else_block(tokens, i)
-
-            if_block.dedent(tokens)
-            del tokens[if_block.end:else_block.end]
-            del tokens[if_block.start:if_block.block]
+            if tokens[i].src == 'if':
+                if_block = Block.find(tokens, i)
+                if_block.dedent(tokens)
+                del tokens[if_block.start:if_block.block]
+            else:
+                if_block = Block.find(tokens, _find_elif(tokens, i))
+                if_block.replace_condition(tokens, [Token('NAME', 'else')])
+        elif token.offset in visitor.if_py2_blocks_else:
+            if tokens[i].src == 'if':
+                if_block, else_block = _find_if_else_block(tokens, i)
+                else_block.dedent(tokens)
+                del tokens[if_block.start:else_block.block]
+            else:
+                j = _find_elif(tokens, i)
+                if_block, else_block = _find_if_else_block(tokens, j)
+                del tokens[if_block.start:else_block.start]
+        elif token.offset in visitor.if_py3_blocks_else:
+            if tokens[i].src == 'if':
+                if_block, else_block = _find_if_else_block(tokens, i)
+                if_block.dedent(tokens)
+                del tokens[if_block.end:else_block.end]
+                del tokens[if_block.start:if_block.block]
+            else:
+                j = _find_elif(tokens, i)
+                if_block, else_block = _find_if_else_block(tokens, j)
+                del tokens[if_block.end:else_block.end]
+                if_block.replace_condition(tokens, [Token('NAME', 'else')])
         elif token.offset in visitor.six_type_ctx:
             _replace(i, SIX_TYPE_CTX_ATTRS, visitor.six_type_ctx[token.offset])
         elif token.offset in visitor.six_simple:
